/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2002 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   A canvas implementation.
 */

#include <gtk--/style.h>

#include <algorithm> // STL algorithms
#include <functional>

#include "plugins-gtk/Canvas.hh"

using namespace std;
using namespace GSML;
using namespace GSMP;

/******************
 * GsmpCanvasItem *
 ******************/

GsmpCanvasItem::GsmpCanvasItem (GsmpCanvas* i_canvas)
{
  m_canvas = i_canvas;
  
  m_transparent = false;
  m_layer = 0;
  
  m_canvas -> Add (this);
}

GsmpCanvasItem::~GsmpCanvasItem ()
{
  m_canvas -> Remove (this);
}

/**************
 * GsmpCanvas *
 **************/

GsmpCanvas::GsmpCanvas ()
{
  m_zoom = 1;
  set_events (GDK_EXPOSURE_MASK
	      //| GDK_LEAVE_NOTIFY_MASK
	      | GDK_POINTER_MOTION_MASK
	      | GDK_BUTTON_PRESS_MASK
	      | GDK_BUTTON_RELEASE_MASK
	      //| GDK_POINTER_MOTION_HINT_MASK
	      | GDK_KEY_PRESS_MASK
	      | GDK_KEY_RELEASE_MASK);
  
  if (!m_fonts.Initialize () ) {
    cout << "Cannot initialize font subsystem -> this sucks ..." << endl;
  }
  
  m_fonts.Debug ();
  m_face = m_fonts.NewFace ();
  
  if (!m_face->LoadFace ("Luxi Sans", "Regular") )
    if (!m_face->LoadFace ("Luxi Mono", "Oblique") )
      if (!m_face->LoadFace ("Luxi Mono", "Regular") )
	if (!m_face->LoadFace ("LuciduxSans", "Sans") )
	  if (!m_face->LoadFace ("LuciduxSans", "Sans Oblique") )
	    cout << "No standard Sans font found ..." << endl;
  
  gdk_rgb_init ();
}

GsmpCanvas::~GsmpCanvas ()
{
  delete m_face;
}

const canvas_area& GsmpCanvas::Viewport () const
{
  return m_viewport;
}


void GsmpCanvas::SetViewport (const canvas_point& p)
{
  // TODO: do optimized scrolling of pixmap data
  
  m_viewport.x = p.x;
  m_viewport.y = p.y;
  
  // Queue to rerender the viewport
  
  RenderRequest (m_viewport);
  
  SignalViewportModified (m_viewport);
}

const canvas_area& GsmpCanvas::CanvasArea () const
{
  return m_area;
}

void GsmpCanvas::SetZoom (int n_zoom)
{
  cout << "GsmpCanvas::Zoom: " << n_zoom << endl;
  m_zoom = n_zoom;
  for_each (m_items.begin (), m_items.end (),
	    bind2nd (mem_fun (&GsmpCanvasItem::SetZoom), m_zoom) );
  
  SignalZoomChanged (m_zoom);
}

int GsmpCanvas::Zoom () const
{
  return m_zoom;
}

void GsmpCanvas::GuiRefresh ()
{
  Render ();
}

bool GsmpCanvas::Add (GsmpCanvasItem* item)
{
  // is it already our item?
  deque<GsmpCanvasItem*>::iterator i;
  for (i = m_items.begin (); i != m_items.end (); ++ i) {
    if (*i == item) {
      cout << "Warning: item already added!" << endl;
      return false;
    }
  }
  m_items.push_back (item);
  
  ItemAreaEnlarged (item -> m_area);
  
  return true;
}

bool GsmpCanvas::Remove (GsmpCanvasItem* item)
{
  // search and remove
  deque<GsmpCanvasItem*>::iterator i;
  for (i = m_items.begin (); i != m_items.end (); ++ i) {
    if (*i == item) {
      canvas_area a = (*i)->m_area;
      m_items.erase (i);
      ItemAreaChanged (a, a);
      return true;
    }
  }
  cout << "Warning: item is not our?!" << endl;
  return false;
}

void GsmpCanvas::RenderRequest (const canvas_area& a)
{
  // for now simply add to the render queue
  m_render_queue.push_back (a);
}

void GsmpCanvas::ItemAreaEnlarged (const canvas_area& item_area)
{
  canvas_area t_area = m_area;
  t_area.Extend (item_area);
  
  if ( !(m_area == t_area) ) {
    m_area = t_area;
    SignalAreaResized (m_area);
  }
  
  RenderRequest (item_area);
}

void GsmpCanvas::ItemAreaChanged (const canvas_area& item_area, const canvas_area& old_item_area)
{
  canvas_area t_area;
  
  /* one complete search wether the whole canvas area is now smaller */
  deque<GsmpCanvasItem*>::iterator i;
  for (i = m_items.begin (); i != m_items.end (); ++ i)
    t_area.Extend ( (*i)->m_area);
  
  if ( !(t_area == m_area) ) {
    m_area = t_area;
    SignalAreaResized (m_area);
  }
  
  RenderRequest (item_area);
}

void GsmpCanvas::Render ()
{
  canvas_area render_area;
  bool render_added = false;
  
  // clip all request to the viewport
  while (m_render_queue.size () > 0)
    {
      canvas_area a = m_render_queue.back ();
      
      cout << "  render request: " << a.x << endl;
      
      a.Clip (m_viewport);
      
      if (a.Value () > 0) {
	if (!render_added) {
	  render_area = a;
	  render_added = true;
	}
	else
	  render_area.Extend (a);
      }
	
      m_render_queue.pop_back ();
    }
  
  // if s.th. to render was found - render all items in this region
  if (render_added && render_area.Value () > 0)
    {
      cout << "  merged r_area: " << render_area << endl;
      
      // calculate the area on the screen (needed for expose below, too)
      GsmpPixmap::area update_rect;
      update_rect.x = render_area.x - m_viewport.x;
      update_rect.y = render_area.y - m_viewport.y;
      update_rect.w = render_area.w;
      update_rect.h =  render_area.h;
      
      cout << "  update_rect (in pixmap): " << update_rect << endl;
      
      // clear the the area we have to render
      m_pixmap.DrawFilledRectangle (update_rect, 0xffffffff);
      
      // render for all items
      deque<GsmpCanvasItem*>::iterator i;
      for (i = m_items.begin (); i != m_items.end (); ++ i)
	{
	  canvas_area item_area = (*i) -> m_area;
	  cout << "  item area: " << item_area << endl;
	  
	  canvas_area item_render_area = render_area;
	  item_render_area.Clip (item_area);
	  
	  // s.th. of the item visible ??
	  if (item_render_area.Value () > 0)
	    {
	      // calculate the dest in the pixmap
	      canvas_point t_dest;
	      t_dest.x = item_render_area.x - m_viewport.x;
	      t_dest.y = item_render_area.y - m_viewport.y;
	      
	      // transform to item coordinates
	      item_render_area.x -= item_area.x;
	      item_render_area.y -= item_area.y;
	      
	      cout << "  clipped for item: " << item_render_area << endl;
	      
	      // debug - draw a rectange arround the area we have to render into
	      //GsmpPixmap::area t_rect (t_dest.x, t_dest.y,
	      //		         item_render_area.w, item_render_area.h);
	      //m_pixmap.DrawRectangle (t_rect, 0xff00a0ff);
	      
	      (*i) -> Render (&m_pixmap, item_render_area, t_dest);
	    }
	  else
	    cout << "  Item not visible here" << endl;
	}
      
      // debug - draw a rectange arround the area we have to rendered into
      m_pixmap.DrawRectangle (update_rect, 0x8000b0ff);
      
      GdkRectangle t_rect;
      t_rect.x = update_rect.x;
      t_rect.y = update_rect.y;
      t_rect.width = update_rect.w;
      t_rect.height = update_rect.h;
      
      cout << "Gdk expose area: " << update_rect << endl;
      
      draw (&t_rect);
    }
  /*
  else
    cout << "Nothing to render ..." << endl;
  */
}

gint GsmpCanvas::configure_event_impl (GdkEventConfigure* event)
{
  cout << "GsmpCanvas::configure_event_impl ..." << endl;
  Gtk::DrawingArea::configure_event_impl (event);
  
  m_win = get_window ();
  if (!m_win)
    cout << "NormalTrackView::configure_event_impl: Window not vaild?" << endl;
  
  m_viewport.w = width ();
  m_viewport.h = height ();
  
  m_pixmap.Resize (GsmpPixmap::point (m_viewport.w, m_viewport.h) );
  
  // queue to rerender the viewport
  RenderRequest (m_viewport);
  
  SignalViewportModified (m_viewport);
  
  return true;
}

gint GsmpCanvas::expose_event_impl (GdkEventExpose* event)
{
  if (m_pixmap.m_data != 0)
    {
      /*
	m_pixmap.DrawLine (GsmpPixmap::point (0, 10),
	GsmpPixmap::point (m_pixmap.Size ().x - 1, 10), 0xffff0000);
	m_pixmap.DrawLine (GsmpPixmap::point (10, 0),
	GsmpPixmap::point (10, m_pixmap.Size ().y - 1), 0xffff0000);
	m_pixmap.DrawLine (GsmpPixmap::point (50, 0),
	GsmpPixmap::point (50, m_pixmap.Size ().y - 1), 0xffff0000);
	m_pixmap.DrawLine (GsmpPixmap::point (100, 0),
	GsmpPixmap::point (100, m_pixmap.Size ().y - 1), 0xffff0000);
      */
      
      GsmpPixmap::area expose_area (event->area.x, event->area.y,
				    event->area.width, event->area.height);
      GsmpPixmap::area t_area (expose_area);
      
      expose_area.Clip (m_pixmap.Area () );
      
      if (expose_area != t_area) {
	cout << "GsmpCanvas::expose_event_impl: BUG: exopse out of range !!!" << endl;
	cout << "  was: " << t_area << ", corrected to: " << expose_area << endl;
	cout << "  pixmap: " << m_pixmap.Area () << endl;
      }
      
      unsigned char* buffer
	= (unsigned char*) (m_pixmap.m_data + event->area.y * m_pixmap.Size ().x + event->area.x);
      
      gdk_draw_rgb_32_image (this->get_window (),
			     this->get_style () -> get_fg_gc (GTK_STATE_NORMAL),
			     expose_area.x,
			     expose_area.y,
			     expose_area.w,
			     expose_area.h,
			     GDK_RGB_DITHER_MAX,
			     buffer,
			     m_pixmap.Size ().x * 4); // rowstride (in bytes)
    }
  else
    cout << "no vaild pixmap yet ..." << endl;
  
  return true;
}

gint GsmpCanvas::motion_notify_event_impl (GdkEventMotion* event)
{
  cout << "GsmpCanvas::motion_notify_event_impl: "
       << "  x: " << (int) event->x
       << ", y: " << (int) event->y
       << ", state: " << event->state << endl;
  
  m_pointer.x = m_viewport.x + (canvas_coord) event->x;
  m_pointer.y = m_viewport.y + (canvas_coord) event->y;
  
  SignalMotionNotify (m_pointer);
  
  /*
  // BitBlt test:
  m_pixmap.BitBlt (GsmpPixmap::area (0, 0, 20, 20),
		   GsmpPixmap::point (event->x, event->y) );
  
  // another BitBlt test:
  GsmpPixmap t;
  t.Resize (GsmpPixmap::point (50, 50) );
  t.Clear (0x20);
  t.DrawLine (GsmpPixmap::point (0,9), GsmpPixmap::point (9,9), 0x80ff2020);
  t.DrawLine (GsmpPixmap::point (9,0), GsmpPixmap::point (9,9), 0x80ff2020);
  
  m_pixmap.BitBlt (t, GsmpPixmap::area (0, 0, 20, 20),
	 	   GsmpPixmap::point ( (int) event->x, (int) event->y) );
  */
  
  /*
  GdkRectangle t_rect;
  t_rect.x = event->x;
  t_rect.y = event->y;
  t_rect.width = 20;
  t_rect.height = 20;
  draw (&t_rect);
  */
  
  GsmpPixmap::point t_mid (m_pixmap.Size ().x/2, m_pixmap.Size ().y/2);
  GsmpPixmap::point t_pointer (m_pointer.x - m_viewport.x, m_pointer.y - m_viewport.y);
  
  m_pixmap.DrawAaLine (t_mid, t_pointer, m_c);
  
  GdkRectangle t_rect;
  
  t_rect.x = std::min (t_mid.x, t_pointer.x);
  t_rect.y = std::min (t_mid.y, t_pointer.y);
  t_rect.width = std::max (t_mid.x, t_pointer.x) - t_rect.x + 1;
  t_rect.height = std::max (t_mid.y, t_pointer.y) - t_rect.y + 1;
  
  draw (&t_rect);
  
  
  
  // FreeType2 test
  
  m_face->Test (&m_pixmap);
  
  t_rect.x = 0;
  t_rect.y = 0;
  t_rect.width = 80;
  t_rect.height = 40;
  
  draw (&t_rect);
  
  return true;
}

gint GsmpCanvas::button_press_event_impl (GdkEventButton* event)
{
  cout << "GsmpCanvas::button_press_event_impl" << endl;
  
  GsmpColor::Component r, g, b, a;
  GsmpColor::Rgba (m_c, &r, &g, &b, &a);
  r += 10;
  g += 20;
  b += 30;
  a += 5;
  GsmpColor::SetRgba (&m_c, r, g, b, a);
  
  // TODO: pass through items
  SignalButtonPress (event->button);
  return true;
}

gint GsmpCanvas::button_release_event_impl (GdkEventButton* event)
{
  cout << "GsmpCanvas::button_release_event_impl" << endl;
  // TODO: pass through items
  SignalButtonRelease (event->button);
  return true;
}

gint GsmpCanvas::key_press_event_impl (GdkEventKey* event)
{
  cout << "GsmpCanvas::key_press_event_impl: keyval: " << event->keyval << endl;
  // TODO: pass through items
  SignalKeyPress (event->keyval);
  return true;
}

gint GsmpCanvas::key_release_event_impl (GdkEventKey* event)
{
  cout << "GsmpCanvas::key_release_event_impl" << endl;
  // TODO: pass through items
  SignalKeyRelease (event->keyval);
  return true;
}
